"""
Conventions for canonical CRUD endpoints.

"""
from flask import jsonify
from inflection import pluralize

from microcosm_flask.conventions.encoding import (
    dump_response_data,
    load_query_string_data,
    load_request_data,
    merge_data,
    require_response_data,
)
from microcosm_flask.conventions.registry import qs, request, response
from microcosm_flask.naming import collection_path_for, instance_path_for, name_for
from microcosm_flask.operations import Operation
from microcosm_flask.paging import Page, PaginatedList, make_paginated_list_schema


# local registry of CRUD mappings
CRUD_MAPPINGS = {}


def _crud(operation):
    """
    Register a "register_<foo>_endpoint" function with an operation.

    """
    def decorator(func):
        CRUD_MAPPINGS[operation] = func
        return func
    return decorator


@_crud(Operation.Search)
def register_search_endpoint(graph, obj, path_prefix, func, request_schema, response_schema):
    """
    Register a search endpoint.

    :param func: a search function, which must:
      - accept kwargs for the query string (minimally for pagination)
      - return a tuple of (items, count) where count is the total number of items
        available (in the case of pagination)
    :param request_schema: a marshmallow schema to decode/validate query string arguments
    :param response_schema: a marshmallow schema to encode (a single) response item
    """

    paginated_list_schema = make_paginated_list_schema(obj, response_schema)()

    @graph.route(path_prefix + collection_path_for(obj), Operation.Search, obj)
    @qs(request_schema)
    @response(paginated_list_schema)
    def search(**path_data):
        request_data = load_query_string_data(request_schema)
        page = Page.from_query_string(request_data)
        items, count = func(**merge_data(path_data, request_data))
        # TODO: use the schema for encoding
        return jsonify(
            PaginatedList(obj, page, items, count, response_schema).to_dict()
        )

    search.__doc__ = "Search the collection of all {}".format(pluralize(name_for(obj)))


@_crud(Operation.Create)
def register_create_endpoint(graph, obj, path_prefix, func, request_schema, response_schema):
    """
    Register a create endpoint.

    :param func: a create function, which must:
      - accept kwargs for the request and path data
      - return a new item
    :param request_schema: a marshmallow schema to decode/validate request data
    :param response_schema: a marshmallow schema to encode response data

    """
    @graph.route(path_prefix + collection_path_for(obj), Operation.Create, obj)
    @request(request_schema)
    @response(response_schema)
    def create(**path_data):
        request_data = load_request_data(request_schema)
        response_data = func(**merge_data(path_data, request_data))
        return dump_response_data(response_schema, response_data, Operation.Create.value.default_code)

    create.__doc__ = "Create a new {}".format(name_for(obj))


@_crud(Operation.Retrieve)
def register_retrieve_endpoint(graph, obj, path_prefix, func, response_schema):
    """
    Register a retrieve endpoint.

    :param func: a retrieve function, which must:
      - accept kwargs for path data
      - return an item or falsey
    :param response_schema: a marshmallow schema to encode response data

    """
    @graph.route(path_prefix + instance_path_for(obj), Operation.Retrieve, obj)
    @response(response_schema)
    def retrieve(**path_data):
        response_data = require_response_data(func(**path_data))
        return dump_response_data(response_schema, response_data)

    retrieve.__doc__ = "Retrieve a {} by id".format(name_for(obj))


@_crud(Operation.Delete)
def register_delete_endpoint(graph, obj, path_prefix, func):
    """
    Register a delete endpoint.

    :param func: a delete function, which must:
      - accept kwargs for path data
      - return truthy/falsey

    """
    @graph.route(path_prefix + instance_path_for(obj), Operation.Delete, obj)
    def delete(**path_data):
        require_response_data(func(**path_data))
        return "", Operation.Delete.value.default_code

    delete.__doc__ = "Delete a {} by id".format(name_for(obj))


@_crud(Operation.Replace)
def register_replace_endpoint(graph, obj, path_prefix, func, request_schema, response_schema):
    """
    Register a replace endpoint.

    :param func: a replace function, which must:
      - accept kwargs for the request and path data
      - return the replaced item
    :param request_schema: a marshmallow schema to decode/validate request data
    :param response_schema: a marshmallow schema to encode response data

    """
    @graph.route(path_prefix + instance_path_for(obj), Operation.Replace, obj)
    @request(request_schema)
    @response(response_schema)
    def replace(**path_data):
        request_data = load_request_data(request_schema)
        # Replace/put should create a resource if not already present, but we do not
        # enforce these semantics at the HTTP layer. If `func` returns falsey, we
        # will raise a 404.
        response_data = require_response_data(func(**merge_data(path_data, request_data)))
        return dump_response_data(response_schema, response_data)

    replace.__doc__ = "Create or update a {} by id".format(name_for(obj))


@_crud(Operation.Update)
def register_update_endpoint(graph, obj, path_prefix, func, request_schema, response_schema):
    """
    Register an update endpoint.

    :param func: an update function, which must:
      - accept kwargs for the request and path data
      - return an updated item
    :param request_schema: a marshmallow schema to decode/validate request data
    :param response_schema: a marshmallow schema to encode response data

    """
    @graph.route(path_prefix + instance_path_for(obj), Operation.Update, obj)
    @request(request_schema)
    @response(response_schema)
    def update(**path_data):
        # NB: using partial here means that marshmallow will not validate required fields
        request_data = load_request_data(request_schema, partial=True)
        response_data = require_response_data(func(**merge_data(path_data, request_data)))
        return dump_response_data(response_schema, response_data)

    update.__doc__ = "Update some or all of a {} by id".format(name_for(obj))


def configure_crud(graph, obj, mappings, path_prefix=""):
    """
    Register CRUD endpoints for a resource object.

    :param mappings: a dictionary from operations to tuple, where each tuple contains
                     the target function and zero or more marshmallow schemas according
                     to the signature of the "register_<foo>_endpoint" functions

    Example mapping:

        {
            Operation.Create: (create_foo, NewFooSchema(), FooSchema()),
            Operation.Delete: (delete_foo,),
            Operation.Retrieve: (retrieve_foo, FooSchema()),
        }

    """
    for operation, register in CRUD_MAPPINGS.items():
        if operation in mappings:
            register(graph, obj, path_prefix, *mappings[operation])



"""
Conventions for canonical relation endpoints (mapping one resource to another).

For relations, operation definitions require *two* objects instead of one; for
simplicity these are passed as a pair/tuple.

"""
from flask import jsonify
from inflection import pluralize

from microcosm_flask.conventions.encoding import (
    dump_response_data,
    load_query_string_data,
    load_request_data,
    merge_data,
)
from microcosm_flask.conventions.registry import qs, request, response
from microcosm_flask.naming import name_for, relation_path_for
from microcosm_flask.operations import Operation
from microcosm_flask.paging import Page, PaginatedList, make_paginated_list_schema


# local registry of relation mappings
RELATION_MAPPINGS = {}


def _relation(operation):
    """
    Register a "register_<foo>_relation_endpoint" function with an operation.

    """
    def decorator(func):
        RELATION_MAPPINGS[operation] = func
        return func
    return decorator


@_relation(Operation.CreateFor)
def register_createfor_relation_endpoint(graph, obj, path_prefix, func, request_schema, response_schema):
    """
    Register a create-for relation endpoint.

    :param func: a store create function, which must:
      - accept kwargs for the new instance creation parameters
      - return the created instance

    :param request_schema: a marshmallow schema to decode/validate instance creation parameters
    :param response_schema: a marshmallow schema to encode the created instance
    """
    @graph.route(path_prefix + relation_path_for(*obj), Operation.CreateFor, obj)
    @request(request_schema)
    @response(response_schema)
    def create(**path_data):
        request_data = load_request_data(request_schema)
        response_data = func(**merge_data(path_data, request_data))
        return dump_response_data(response_schema, response_data, Operation.CreateFor.value.default_code)

    create.__doc__ = "Create a new {} relative to a {}".format(pluralize(name_for(obj[1])), name_for(obj[0]))


@_relation(Operation.SearchFor)
def register_search_relation_endpoint(graph, obj, path_prefix, func, request_schema, response_schema):
    """
    Register a relation endpoint.

    :param func: a search function, which must:
      - accept kwargs for the query string (minimally for pagination)
      - return a tuple of (items, count, context) where count is the total number of items
        available (in the case of pagination) and context is a dictionary providing any
        needed context variables for constructing pagination links

    :param request_schema: a marshmallow schema to decode/validate query string arguments
    :param response_schema: a marshmallow schema to encode (a single) response item
    """

    paginated_list_schema = make_paginated_list_schema(obj[1], response_schema)()

    @graph.route(path_prefix + relation_path_for(*obj), Operation.SearchFor, obj)
    @qs(request_schema)
    @response(paginated_list_schema)
    def search(**path_data):
        request_data = load_query_string_data(request_schema)
        page = Page.from_query_string(request_data)
        items, count, context = func(**merge_data(path_data, request_data))
        # TODO: use the schema for encoding
        return jsonify(
            PaginatedList(obj, page, items, count, response_schema, Operation.SearchFor, **context).to_dict()
        )

    search.__doc__ = "Search for {} relative to a {}".format(pluralize(name_for(obj[1])), name_for(obj[0]))


def configure_relation(graph, from_obj, to_obj, mappings, path_prefix=""):
    """
    Register relation endpoint(s) between two resources.

    """
    for operation, register in RELATION_MAPPINGS.items():
        if operation in mappings:
            register(graph, (from_obj, to_obj), path_prefix, *mappings[operation])



"""
A naming convention and discovery mechanism for HTTP endpoints.

Operations provide a naming convention for references between endpoints,
allowing easy construction of links or audit trails for external consumption.

"""
from collections import namedtuple
from urllib import urlencode
from urlparse import urljoin

from enum import Enum, unique
from flask import request, url_for

from microcosm_flask.naming import name_for


# metadata for an operation
OperationInfo = namedtuple("OperationInfo", ["name", "method", "pattern", "default_code"])


NODE_PATTERN = "{}.{}"
EDGE_PATTERN = "{}.{}.{}"


@unique
class Operation(Enum):
    """
    An enumerated set of operation types, which know how to resolve themselves into
    URLs and hrefs.

    """
    # discovery operation
    Discover = OperationInfo("discover", "GET", NODE_PATTERN, 200)

    # collection operations
    Search = OperationInfo("search", "GET", NODE_PATTERN, 200)
    Create = OperationInfo("create", "POST", NODE_PATTERN, 201)
    # bulk update is possible here with PATCH

    # instance operations
    Retrieve = OperationInfo("retrieve", "GET", NODE_PATTERN, 200)
    Delete = OperationInfo("delete", "DELETE", NODE_PATTERN, 204)
    Replace = OperationInfo("replace", "PUT", NODE_PATTERN, 200)
    Update = OperationInfo("update", "PATCH", NODE_PATTERN, 200)

    # relation operations
    CreateFor = OperationInfo("create_for", "POST", EDGE_PATTERN, 201)
    SearchFor = OperationInfo("search_for", "GET", EDGE_PATTERN, 200)

    # ad hoc operations
    Command = OperationInfo("command", "POST", NODE_PATTERN, 200)
    Query = OperationInfo("query", "GET", NODE_PATTERN, 200)

    def name_for(self, obj):
        """
        Generate an operation name in the scope of one or more resources.

        This naming convention matches how Flask blueprints routes are resolved
        (assuming that the blueprint and resources share the same name).

        Examples: `foo.search`, `bar.search_for.baz`

        """
        if isinstance(obj, (list, tuple)):
            return self.value.pattern.format(
                name_for(obj[0]),
                self.value.name,
                name_for(obj[1])
            )
        else:
            return self.value.pattern.format(
                name_for(obj),
                self.value.name,
            )

    def url_for(self, obj, **kwargs):
        """
        Construct a URL for an operation against a resource.

        :param kwargs: additional arguments for URL path expansion

        """
        return url_for(self.name_for(obj), **kwargs)

    def href_for(self, obj, qs=None, **kwargs):
        """
        Construct an full href for an operation against a resource.

        :parm qs: the query string dictionary, if any
        :param kwargs: additional arguments for path expansion
        """
        return "{}{}".format(
            urljoin(request.url_root, self.url_for(obj, **kwargs)),
            "?{}".format(urlencode(qs)) if qs else "",
        )

    @classmethod
    def from_name(cls, name):
        for operation in cls:
            if operation.value.name == name:
                return operation
        else:
            raise ValueError(name)

    @classmethod
    def parse(cls, name):
        """
        Convert an operation name back to an operation, obj tuple.

        """
        parts = name.split(".")
        operation = cls.from_name(parts[1])
        if len(parts) > 2:
            return operation, parts[0:1] + parts[2:]
        else:
            return operation, parts[0]



"""
Push resource definitions to an output destination.

"""
from json import dumps
from logging import getLogger
from sys import stdout
from urlparse import urljoin

from requests import put
from yaml import safe_dump_all


logger = getLogger("sync.push")


def push_yaml(inputs, destination):
    """
    Write inputs to destination as YAML.

    :param inputs: an iterable of (href, resource) pairs
    :param destination: a writable file-like object

    """
    safe_dump_all(({href: resource} for href, resource in inputs), destination)


def push_json(inputs, base_url):
    """
    Write inputs to remote URL as JSON.

    Future implementations could perform paginated PATCH requests to a collection URI
    if our conventions support it.

    """
    for href, resource in inputs:
        # Skip over links-only (discovery) resources
        if resource.keys() == ["_links"]:
            continue

        # Note that `urljoin` will only use the *base* URL of the destination.
        # If we need more control over URI paths, we'll need to use a different function.
        uri = urljoin(base_url, href)
        push_resource_json(uri, resource)


def push_resource_json(uri, resource):
    """
    Push a single resource as JSON to a URI.

    Assumes that the backend supports a replace/put convention.

    """
    # We don't want to submit links back; they are synthetic
    logger.debug("Pushing resource for {}".format(uri))

    response = put(
        uri,
        data=dumps(resource),
        headers={"Content-Type": "application/json"},
    )
    try:
        response.raise_for_status()
    except:
        logger.warn("Unable to replace {}".format(
            uri,
        ))
        raise


def push(args, inputs):
    """
    Push content to a destination.

    If the destination is "-", YAML is written to stdout.
    If the destination has a http prefix, JSON is written to a URL.
    Otherwise, YAML is written to a local file.

    """
    logger.info("Pushing resources to: {}".format(args.output))

    if args.output == "-":
        push_yaml(inputs, stdout)
    elif args.output.startswith("http"):
        push_json(inputs, args.output)
    else:
        with open(args.output, "w") as file_:
            push_yaml(inputs, file_)



"""
Swagger (OpenAPI) convention.

Exposes swagger definitions for matching operations.

"""
from flask import jsonify, g

from microcosm.api import defaults
from microcosm_flask.conventions.registry import iter_operations
from microcosm_flask.naming import singleton_path_for
from microcosm_flask.operations import Operation
from microcosm_flask.swagger.definitions import build_swagger


def register_swagger_endpoint(graph, name, version, path_prefix, match_func):
    """
    Register a swagger endpoint for a set of operations.

    """
    @graph.route(path_prefix + singleton_path_for(name), Operation.Discover, name)
    def discover():
        operations = list(iter_operations(graph, match_func))

        swagger = build_swagger(graph, version, path_prefix, operations)

        g.hide_body = True

        return jsonify(swagger)

    return name


@defaults(
    name="swagger",
    operations=[
        "create",
        "delete",
        "replace",
        "retrieve",
        "search",
        "search_for",
        "update",
    ],
    path_prefix="",
)
def configure_swagger(graph):
    """
    Build a singleton endpoint that provides swagger definitions for all operations.

    """
    name = graph.config.swagger_convention.name
    version = graph.config.swagger_convention.version

    base_path = graph.config.route.path_prefix
    path_prefix = graph.config.swagger_convention.path_prefix + "/" + version

    matches = {
        Operation.from_name(operation_name.lower())
        for operation_name in graph.config.swagger_convention.operations
    }

    def match_func(operation, obj, rule):
        return (
            rule.rule.startswith(base_path + path_prefix) and
            operation in matches
        )

    return register_swagger_endpoint(graph, name, version, path_prefix, match_func)



"""
A discovery endpoint provides links to other endpoints.

"""
from flask import jsonify

from microcosm.api import defaults
from microcosm_flask.conventions.encoding import load_query_string_data
from microcosm_flask.conventions.registry import iter_operations
from microcosm_flask.linking import Link, Links
from microcosm_flask.naming import name_for, singleton_path_for
from microcosm_flask.paging import Page, PageSchema
from microcosm_flask.operations import Operation


def iter_links(operations, page):
    """
    Generate links for an iterable of operations on a starting page.

    """
    for operation, obj, rule, func in operations:
        yield Link.for_(
            operation=operation,
            obj=obj,
            type=name_for(obj),
            qs=page.to_tuples(),
        )


def register_discovery_endpoint(graph, name, match_func, path_prefix=""):
    """
    Register a discovery endpoint for a set of operations.

    """
    page_schema = PageSchema()

    @graph.route(path_prefix + singleton_path_for(name), Operation.Discover, name)
    def discover():
        # accept pagination limit from request
        page = Page.from_query_string(load_query_string_data(page_schema))
        page.offset = 0

        operations = list(iter_operations(graph, match_func))

        return jsonify(
            _links=Links({
                "self": Link.for_(Operation.Discover, name, qs=page.to_tuples()),
                "search": [
                    link for link in iter_links(operations, page)

                ],
            }).to_dict()
        )

    return name


@defaults(
    name="",
    operations=[
        "search",
    ],
    path_prefix="",
)
def configure_discovery(graph):
    """
    Build a singleton endpoint that provides a link to all search endpoints.

    """
    name = graph.config.discovery_convention.name
    path_prefix = graph.config.discovery_convention.path_prefix

    matches = {
        Operation.from_name(operation_name.lower())
        for operation_name in graph.config.discovery_convention.operations
    }

    def match_func(operation, obj, rule):
        return operation in matches

    return register_discovery_endpoint(graph, name, match_func, path_prefix)



"""
Naming conventions for Swagger definitions.

Intended to play nice with generated code (e.g. with Bravado)

"""

from inflection import camelize, pluralize

from microcosm_flask.naming import name_for


def operation_name(operation, obj):
    """
    Convert an operation, obj(s) pair into a swagger operation id.

    For compatability with Bravado, we want to use underscores instead of dots and
    verb-friendly names. Example:

        foo.retrieve       => client.foo.retrieve()
        foo.search_for.bar => client.foo.search_for_bars()

    """
    if isinstance(obj, (list, tuple)):
        verb, rest = operation.value.name, obj[1:]
    else:
        verb, rest = operation.value.name, []
    return "_".join([verb] + [pluralize(name_for(noun)) for noun in rest])


def type_name(name):
    """
    Convert an internal name into a swagger type name.

    For example:

        foo_bar => FooBar

    """
    if name.endswith("_schema"):
        name = name[:-7]
    return camelize(name)
